# Escaped variable references by default:
{{vars.init.variable_containing_double_quote}}
{{vars.init.variable_containing_single_quote}}
# Unescaped variable references:
{{&vars.init.variable_containing_double_quote}}
{{&vars.init.variable_containing_single_quote}}
{{{vars.init.variable_containing_double_quote}}}
{{{vars.init.variable_containing_single_quote}}}
